#!/usr/bin/env python
'''Phase field programm for Cahn-Hillard equation
# Copyright (c) 2013 Benoit appolaire <benoit.appolaire@onera.fr>
# and Yann Le Bouar <yann.lebouar@onera.fr>
# Permission is hereby granted, free of charge, to any person 
# obtaining a copy of this software and associated documentation 
# files (the "Software"), to deal in the Software without restriction, 
# including without limitation the rights to use, copy, modify, merge, 
# publish, distribute, sublicense, and/or sell copies of the Software, 
# and to permit persons to whom the Software is furnished to do so, 
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be 
# included in all copies or substantial portions of the Software.
'''

#______________________________________________________________________________
#								Import modules/libraries
#______________________________________________________________________________

import sys				#system basic routines (may be usefull for debugging)
from scipy import *		#scientific library including fast fourier transform
from pylab import *		#plotting a la matlab with matplotlib
import numpy.random as rng #for random numbers
import operator			#export functions for the intrinsic operators


#______________________________________________________________________________
#							Define a few functions/subroutines
#______________________________________________________________________________



#..............................................................................
def Egrad(phi_,kx_,ky_,alpha):
	'''Gradient energy density'''
#..............................................................................
	tmp = 0.5*alpha*(kx_**2+ky_**2)*abs(phi_)**2
	return tmp.real
	
#..............................................................................
def g(phi):
	'''Double well potential where the energy barrier is 1'''
#..............................................................................

	return 16.*phi**2.*(1.-phi)**2.
	
#..............................................................................
def dg(phi):
	'''Derivative of the double well potential $g(\phi) = 16*\phi^2(1-\phi)^2$
		i.e. $\frac{dg}{dphi}$ (Non linear source term in the pde)'''
#..............................................................................

	return 2.*phi*(1.-phi)*(1.-2.*phi)*16.

#..............................................................................
def ddg(phi):
	'''Second Derivative of the double well potential $g(\phi) = 16\phi^2(1-\phi)^2$
		i.e. $\frac{d^2g}{dphi^2}$'''
#..............................................................................

	return 32.*(1.-6.*phi*(1.-phi))

#..............................................................................
def discretization(L=1.,N=10):
	'''Discretize with N nodes the real and reciprocal spaces 
		along one direction with length 2*L 
	'''
#..............................................................................

	return mgrid[-L:L:2*L/N], fftfreq(N,d=L/N)

#..............................................................................
def init_cosx(x_,y_,scalex=1.):
	'''Initialization of a field with cosine along x
	'''
#..............................................................................
	
	return (1.+cos(2.*pi*x_/scalex))/2.

#..............................................................................
def init_circle(x_,y_,centerx=0.,centery=0.,radius=1.,val1=1.,val2=0.):
	'''Initialization of a field with a disk centered at (centerx,centery)
		and with a given radius
	'''
	
#..............................................................................
	
	return where( (x_-centerx)**2.+(y_-centery)**2.-radius**2.<=0., val1, val2 )

#..............................................................................
def init_interface(x_,y_,pos1=0,pos2=1,val1=1.,val2=0.):
	'''Initialization of a field with two interfaces normal to x
	'''
	
#..............................................................................
	
	return where( (( x_ >= pos1) & ( x_ < pos2)), val1, val2 )

#..............................................................................
def iterwhite():
	'''Generate a sequence of samples of white noise.
		Generates a never-ending sequence of floating-point values.
		
		# Copyright (c) 2011 Leif Johnson <leif@leifjohnson.net>
		#
		# Permission is hereby granted, free of charge, to any person 
		# obtaining a copy of this software and associated documentation 
		# files (the "Software"), to deal in the Software without restriction, 
		# including without limitation the rights to use, copy, modify, merge, 
		# publish, distribute, sublicense, and/or sell copies of the Software, 
		# and to permit persons to whom the Software is furnished to do so, 
		# subject to the following conditions:
		#
		# The above copyright notice and this permission notice shall be 
		# included in all copies or substantial portions of the Software.
	'''
#..............................................................................

	while True:
		for n in rng.randn(100):
			yield n

#..............................................................................
def _asarray(source, shape):
	'''
		# Copyright (c) 2011 Leif Johnson <leif@leifjohnson.net>
		#
		# Permission is hereby granted, free of charge, to any person 
		# obtaining a copy of this software and associated documentation 
		# files (the "Software"), to deal in the Software without restriction, 
		# including without limitation the rights to use, copy, modify, merge, 
		# publish, distribute, sublicense, and/or sell copies of the Software, 
		# and to permit persons to whom the Software is furnished to do so, 
		# subject to the following conditions:
		#
		# The above copyright notice and this permission notice shall be 
		# included in all copies or substantial portions of the Software.

	'''
#..............................................................................

	noise = source()
	if shape is None:
		return noise.next()
	count = reduce(operator.mul, shape)
	return asarray([noise.next() for _ in range(count)]).reshape(shape)

#..............................................................................
def white(shape=None):
	'''Generate white noise.

		shape: If given, returns a numpy array of white noise with this shape. If
		not given, return just one sample of white noise.

		# Copyright (c) 2011 Leif Johnson <leif@leifjohnson.net>
		#
		# Permission is hereby granted, free of charge, to any person 
		# obtaining a copy of this software and associated documentation 
		# files (the "Software"), to deal in the Software without restriction, 
		# including without limitation the rights to use, copy, modify, merge, 
		# publish, distribute, sublicense, and/or sell copies of the Software, 
		# and to permit persons to whom the Software is furnished to do so, 
		# subject to the following conditions:
		#
		# The above copyright notice and this permission notice shall be 
		# included in all copies or substantial portions of the Software.
	'''
#..............................................................................

	return _asarray(iterwhite, shape)



#______________________________________________________________________________
#		This is the main program which will be run when invoking the script

if __name__ == "__main__":
#______________________________________________________________________________


	#....................................................
	#(0) A few data

	#Phase field gradient (adimensional)
	alpha	  = 18.
	
	#Kinetic coefficient (adimensional)
	kin       = 1./ddg(0)
	
	print 'kin = ', kin

	#Flags to plot, save, store ...
	fmap	  = True
	fcntr	  = False
	fstore    = True
	fsave	  = False
	fenergy	  = True


	#....................................................
	#(1) Space discretization
	nmax    = 100
	xmax    = 1.*nmax

	x,kx	  = discretization(L=xmax,N=nmax)
	y,ky	  = discretization(L=xmax,N=nmax)
	
	x_,y_   = meshgrid(x,y)
	kx_,ky_ = meshgrid(kx,ky)

	kx_ = 2*pi*kx_
	ky_ = 2*pi*ky_


	#....................................................
	#(2) Time discretization
	tmax   = 20.
	pdt    = 1e-3
	nsteps = int(tmax/pdt)
	steps  = range(1,nsteps+1)

	#	Times when to plot
	#tplots = [80,82,84,86,88,90,92,94,96,98,100,tmax]
	nplots = 10
	tplots = linspace(0.,tmax,nplots+1)
	iplot  = 0


	#....................................................
	#(3) Initial microstructure
	
	phi0   = zeros(shape(x_))

	#phi0   += init_interface(x_,y_,pos1=-x.max()/2.,pos2=x.max()/2.,val1=1.,val2=0.)
	phi0   += init_circle(x_,y_,centerx=(0.),centery=(0.),radius=(x.max()/4.), val1=1., val2=0.)
	#phi0   += init_circle(x_,y_,centerx=(x.max()/2.),centery=(x.max()/6.),radius=(x.max()/3.), val1=1., val2=0.)
	
	#phi0 = 0.5 + 0.01 * white(shape(x_))
	
	#phi0 += 0.01 * white(shape(x_))

	if fstore:
		savez('phi%02d'%iplot,phi=phi0)


	#....................................................
	#(4) Prepare the plots
	
	nc 	 = 5
	nl 	 = (len(tplots)-1)/nc + 1

	fig1   = figure(1,figsize=(14,3*nl))

	subplot(nl,nc,iplot+1)
	iplot += 1
	if fmap:
		imshow(phi0,interpolation='nearest', \
					origin='lower',cmap=cm.gray,\
									vmin=0.,vmax=1.)
	if fcntr:
		contour(phi0,levels=[0.5],colors='r')
	title('t = 0')
	xticks([])
	yticks([])

	isteps = [0]


	#....................................................
	#(5) Time integration of the PDE

	phi 	 = phi0
	phi_     = fft2(phi)
	grad     = zeros(shape(x_))
	fhom     = zeros(shape(x_))
	phi_av = [average(phi)]
	phi_min= [phi.real.min()]
	phi_max= [phi.real.max()]
	
	print 'Starting the time loop'
	
	for step in steps:
		
		# (5a) Solve using fft
		phi0 = phi
		dg_  = fft2(dg(phi))
		phi_ = fft2(phi)
	
		#---------------------  Allen-Cahn (explicit) ----------------------

		phi_ = phi_ - kin*pdt*(dg_+(kx_**2.+ky_**2.)*alpha*phi_)
		
		#-------------------------------------------------------------------
				
		phi  = ifft2(phi_)

		# (That's all !!!!!)
		
	
		# (5b) Compute quantities for the time evolution plot
		phi_av.append(average(phi.real))
		phi_min.append(phi.real.min())
		phi_max.append(phi.real.max())

		# (5c) Snapshots of the evolution
		if step==int(tplots[iplot]/pdt):
		
			if fenergy:
				fhom = g(phi)
				grad = Egrad(phi_,kx_,ky_,alpha)
				E_grad = average(grad)
				E_homo = nmax*nmax*average(fhom)
				E_tot  = E_grad + E_homo
				print 'time %f:   energies hom = %e  grad = %e  tot = %e  '%(step*pdt, E_grad, E_homo, E_tot)
			
			if fstore:
				savez('phi%02d'%iplot,phi=phi.real)
				print '>> storing'
				print 'time %f:  dphi = %e     average = %e'%(step*pdt,(phi-phi0).real.max(),average(phi.real))

			subplot(nl,nc,iplot+1)
			iplot += 1
			if fmap:
				imshow(phi.real,interpolation='nearest',\
								origin='lower',cmap=cm.gray,\
												vmin=0.,vmax=1.)
			if fcntr:
				contour(phi.real,levels=[0.5],colors='r')
			title('t = %.2f'%(step*pdt))
			xticks([])
			yticks([])

			print '>> plotting'
			isteps.append(step)

	subplots_adjust(bottom=0.05,left=0.05,right=0.95,top=0.95,wspace=0.4)


	#....................................................
	#(6) Plot the time evolution of the average and extrema

	fig2 = figure(2)
	clf()
	axes([0.14,0.12,0.82,0.83])

	trange = array([0]+steps)*pdt
#	phi_av = array(phi_av)*100.

	plot(trange,phi_av,'k-',  label='c_avg')
	plot(trange,phi_min,'c-', label='c_max')
	plot(trange,phi_max,'r-', label='c_min')
	legend( loc='upper right' )

	xlabel('time',fontsize=20)
	ylabel(r'$c$',fontsize=20)


	# plot dots corresponding to the snapshots
	for i in range(iplot):
		plot(trange[isteps[i]],phi_av[isteps[i]], 'ko',markersize=10)
		plot(trange[isteps[i]],phi_min[isteps[i]],'co',markersize=10)
		plot(trange[isteps[i]],phi_max[isteps[i]],'ro',markersize=10)

	if fstore:
		savetxt('time_table.txt',trange[isteps])


	#....................................................
	print
	print 'Done'
	print
	
	if fsave:
		fig1.savefig('snapshots.png')
		fig2.savefig('conc_extrema.pdf')
	else:
		show()

